#include <stdio.h>
#include <string.h>
#include <conio.h>

typedef unsigned char  uint8_t;
typedef unsigned short uint16_t;
typedef unsigned long  uint32_t;

uint16_t mpu_base = 0x330;
uint16_t sb_base = 0x220;

uint8_t stack_base;


uint16_t dump_addr;

uint8_t dsp_ver_h;
uint8_t dsp_ver_l;

uint32_t crc6k;
uint32_t crc8k;

uint16_t error = 0;

#define SB_RESET_PORT   (sb_base + 0x6)
#define SB_READ_DATA    (sb_base + 0xA)
#define SB_WRITE_DATA   (sb_base + 0xC)
#define SB_WRITE_STATUS (sb_base + 0xC)
#define SB_READ_STATUS  (sb_base + 0xE)

#define ERROR_SB_RESET        (1)
#define ERROR_SB_SEND         (2)
#define ERROR_SB_RECV         (3)
#define ERROR_MPU_SEND_CMD    (4)
#define ERROR_MPU_SEND_DATA   (5)
#define ERROR_MPU_RECV_DATA   (6)
#define ERROR_MPU_LOOPBACK    (7)
#define ERROR_DSP_UNKNOWN     (8)
#define ERROR_DUMP_FAILED     (9)

char strbuf[256];
uint8_t memdump[256];

uint8_t dump[8192];

FILE *logfile;

void log_open (void) {
//  logfile = fopen ("log.txt", "at");
}

void log_close (void) {
//  if (logfile != NULL) fclose (logfile);
}

void log (const char *str) {
  printf ("%s", str);

  logfile = fopen ("log.txt", "at");
  if (logfile != NULL) fputs (str, logfile);
  if (logfile != NULL) fclose (logfile);
}



static uint8_t mpu_output_full (void) {
  return(inp(mpu_base + 1) & 0x40);
}
static unsigned char mpu_input_empty(void) {
  return(inp(mpu_base + 1) & 0x80);
}
static uint8_t mpu_read (void) {
  uint16_t i;

  for (i=0; i < 0xFFFF; ++i) {
    if (mpu_input_empty() == 0) break;
  }
  if (i == 0xFFFF) { error = ERROR_MPU_RECV_DATA; return 0; }
  return inp(mpu_base);
}
static void mpu_write (uint8_t byte) {
  uint16_t i;

  for (i=0; i < 0xFFFF; ++i) {
    if (mpu_output_full() == 0) break;
  }
  if (i == 0xFFFF) { error = ERROR_MPU_SEND_DATA; return; }
  outp (mpu_base, byte);
}
static void mpu_command (uint8_t cmd) {
  uint16_t i;

  for (i=0; i < 0xFFFF; ++i) {
    if (mpu_output_full() == 0) break;
  }
  if (i == 0xFFFF) { error = ERROR_MPU_SEND_CMD; return; }
  outp (mpu_base + 1, cmd);
}


void mpu_reset (void) {
  mpu_command (0xFF);
  mpu_read();
}

void mpu_init_try (void) {
  uint8_t b;
  mpu_command (0xFF);
  if (error) return;
  b = mpu_read();
  if (error) return;
  if (b != 0xFE) {
    error = ERROR_MPU_RECV_DATA;
    return;
  }
  mpu_command (0x3F);
  if (error) return;
  b = mpu_read();
  if (error) return;
  if (b != 0xFE) error = ERROR_MPU_RECV_DATA;

}


void mpu_init (void) {
  error = 0;
  mpu_init_try ();
  if (error == 0) return;

  for (;;) {
    mpu_command (0xFF);
    sb_delay (0xFFFF);
    inp (SB_READ_DATA);
    inp (mpu_base);
    error = 0;
    sb_reset ();
    mpu_init_try ();
    if (error == 0) return;
  }

}

void mpu_loopback_check (void) {
  mpu_write (0xCA);
  if (error) return;
  if (mpu_read() != 0xCA) error = ERROR_MPU_LOOPBACK;
  if (error) return;
  mpu_write (0xFE);
  if (error) return;
  if (mpu_read() != 0xFE) error = ERROR_MPU_LOOPBACK;
  if (error) return;
}


static uint8_t sb_read_data_available (void) {
  return inp(SB_READ_STATUS) & 0x80;
}

static void sb_delay (uint16_t cycles) {  // 1 cycle is at least 480ns
  uint16_t i;
  for (i=0; i<cycles; ++i) inp(SB_WRITE_STATUS);
}

static void sb_send (uint8_t data) {
  uint16_t i;

  for (i=0; i < 0xFFFF; ++i) {
    if ((inp(SB_WRITE_STATUS) & 0x80) == 0x00) break;
  }
  if (i == 0xFFFF) { error = ERROR_SB_SEND; return; }

  outp (SB_WRITE_DATA, data);

}

static uint8_t sb_recv (void) {
  uint16_t i;

  for (i=0; i < 0xFFFF; ++i) {
    if (inp(SB_READ_STATUS) & 0x80) break;
  }
  if (i == 0xFFFF) { error = ERROR_SB_RECV; return 0; }

  return inp (SB_READ_DATA);
}


void sb_reset (void) {
  uint16_t i;

  outp (SB_RESET_PORT, 1);
  inp(SB_READ_DATA);
  sb_delay (7);  // more then 3us
  outp (SB_RESET_PORT, 0);
  for (i=0; i < 0xFFFF; ++i) { // 208 = ~100us
    if (sb_read_data_available()) break;
  }
  if (i == 0xFFFF) { error = ERROR_SB_RESET; return; }

  if (inp(SB_READ_DATA) != 0xAA) error = ERROR_SB_RESET;
}

void sb_init (void) {
  inp (SB_READ_DATA);
  inp (mpu_base);
  mpu_command (0xFF);
}



void sb_get_dsp_ver (void) {
  uint8_t h,l;

  sb_send (0xE1);
  if (error) return;
  dsp_ver_h = sb_recv ();
  if (error) return;
  dsp_ver_l = sb_recv ();
  if (error) return;

}

static uint8_t sb_read_memory (uint8_t addr) {
  sb_send (0xF9);
  if (error) return 0;
  sb_send (addr);
  if (error) return 0;
  return sb_recv ();
}

static void sb_write_memory (uint8_t addr, uint8_t data) {
  sb_send (0xFA);
  if (error) return;
  sb_send (addr);
  if (error) return;
  sb_send (data);
}

void sb_clean_memory (uint8_t data) {
  int addr;

  for (addr = 0x40; addr < 256; ++addr) {
    sb_write_memory (addr, data);
    if (error) return;
  }
}

void sb_dump_memory (uint8_t *ptr) {
  int addr;

  for (addr = 0; addr < 256; ++addr) {
    ptr[addr] = sb_read_memory (addr);
    if (error) return;
  }
}

void sb_save_memdump (uint8_t *ptr, char *filename) {
  FILE *file;
  file = fopen (filename, "wb");
  if (file != NULL) {
    fwrite (ptr, 256, 1, file);
    fclose (file);
  }
}

void memory_dump (void) {
  sb_clean_memory (0xFF);
  if (error) return;
  sb_dump_memory (memdump);
  if (error) return;
  sb_save_memdump (memdump, "MDUMP2.bin");
  sb_clean_memory (0x00);
  if (error) return;
  sb_dump_memory (memdump);
  if (error) return;
  sb_save_memdump (memdump, "MDUMP1.bin");

}

void sb_initial_diag (void) {
  error = 0;
  log ("SB Reset: ");
  sb_reset ();
  if (error) { log ("failed!\n"); return; }
  else log ("done\n");

  log ("DSP version: ");
  sb_get_dsp_ver ();
  if (error) { log ("failed!\n"); return; }
  else {
    sprintf (strbuf, "%d.%02d\n", (int)dsp_ver_h, (int)dsp_ver_l);
    log (strbuf);
  }

  if (dsp_ver_h != 0x04) { error = ERROR_DSP_UNKNOWN; return; }


  log ("MPU-401 init: ");
  mpu_init ();
  if (error) { log ("failed!\n"); return; }
  else log ("done\n");

  log ("MPU-401 loopback check: ");
  mpu_loopback_check();
  if (error) { log ("failed!\n"); return; }
  else log ("done\n");

}


void dsp_404_attack1 (void) {
  uint8_t byte;
  int i;

  error = 0;
  sb_reset ();
  mpu_init ();


  sb_write_memory (0x7F, dump_addr >> 8);
  sb_write_memory (0x7E, dump_addr & 0xFF);

  sb_write_memory (0x7C, 0x0F);
  sb_write_memory (0x7B, 0x97);

  sb_write_memory (0x01, 0x82);

  mpu_write (0x02);

  sb_delay (0xFFFF);

  sb_send (0xE3);

  for (i=0; i<256; ++i) {
    byte = sb_recv ();
    dump[dump_addr++] = byte;
    if (dump_addr >= 8192) return;
    if (byte == 0x00) return;
  }
}

void dsp_405_attack1 (void) {
  uint8_t byte;
  int i;

  error = 0;
  sb_reset ();
  mpu_init ();


  sb_write_memory (0x7F, dump_addr >> 8);
  sb_write_memory (0x7E, dump_addr & 0xFF);

  sb_write_memory (0x7C, 0x10);
  sb_write_memory (0x7B, 0x43);

  sb_write_memory (0x01, 0x82);

  mpu_write (0x02);

  sb_delay (0xFFFF);

  sb_send (0xE3);

  for (i=0; i<256; ++i) {
    byte = sb_recv ();
    dump[dump_addr++] = byte;
    if (dump_addr >= 8192) return;
    if (byte == 0x00) return;
  }
}

void dsp_411_attack1 (void) {
  uint8_t byte;
  int i;

  error = 0;
  sb_reset ();
  mpu_init ();


  sb_write_memory (0xBE, dump_addr >> 8);
  sb_write_memory (0xBD, dump_addr & 0xFF);

  sb_write_memory (0xBB, 0x12);
  sb_write_memory (0xBA, 0x26);

  sb_write_memory (0x01, 0xC2);

  mpu_write (0x04);

  sb_delay (0xFFFF);

  sb_send (0xE3);

  for (i=0; i<256; ++i) {
    byte = sb_recv ();
    dump[dump_addr++] = byte;
    if (dump_addr >= 8192) return;
    if (byte == 0x00) return;
  }
}










void error_print (void) {
  if (error) log ("ERROR: ");

  switch (error) {
    case ERROR_SB_RESET:
    log ("Cannot RESET sound blaster\n");
    break;
    case ERROR_SB_SEND:
    log ("Cannot SEND to sound blaster\n");
    break;
    case ERROR_SB_RECV:
    log ("Cannot RECV from sound blaster\n");
    break;
    case ERROR_DSP_UNKNOWN:
    log ("DSP version not supported\n");
    break;

    case ERROR_MPU_SEND_CMD :
    log ("Cannot SEND cmd to MPU\n");
    break;
    case ERROR_MPU_SEND_DATA:
    log ("Cannot SEND data to MPU\n");
    break;
    case ERROR_MPU_RECV_DATA:
    log ("Cannot RECV from MPU\n");
    break;
    case ERROR_MPU_LOOPBACK :
    log ("Midi loopback not present\n");
    break;

    case ERROR_DUMP_FAILED:
    log ("Dump failed\n");
    break;


    default:
    log ("Success!\n");

  }

}

void print_codepoints_addr (void) {

  uint8_t h,l;

  l = sb_read_memory (stack_base + 1);
  h = sb_read_memory (stack_base + 2);

  sprintf (strbuf, "ADDR1: %04X\n", ((uint16_t)h << 8) + l);
  log (strbuf);

  sb_write_memory (0x01, 0x04);
  sb_write_memory (0x02, stack_base + 1);

  if (dsp_ver_l < 10) mpu_write (0x40 - 0x03);
  else mpu_write (0x80 - 0x03);
  
  l = mpu_read();
  h = mpu_read();
  sprintf (strbuf, "ADDR2: %04X\n", ((uint16_t)h << 8) + l);
  log (strbuf);

  l = mpu_read();
  h = mpu_read();
  sprintf (strbuf, "ADDR3: %04X\n", ((uint16_t)h << 8) + l);
  log (strbuf);

  sb_write_memory (0x01, 0x40);
  sb_write_memory (0x02, 0x40);
  if (dsp_ver_l < 10) sb_write_memory (0x04, 0x40);
  else sb_write_memory (0x04, 0x80);

}

uint8_t stage1_inject1 (uint8_t variant) {
  uint8_t i;
  
  error = 0;
  sb_reset ();
  mpu_init ();

  sb_write_memory (0x01, 0xC4);
  for ( i = stack_base - 0x10; i <= stack_base; ++i) sb_write_memory (i, 0x00);
  mpu_write (variant & 0x03);
  sb_delay (0xFFFF);

  i = sb_recv ();
  if (error != 0) return 0;
  if (i != 0xAA) return 0;
  return 1;
}

uint8_t stage1_inject2 (uint8_t variant) {
  uint8_t i;
  
  error = 0;
  sb_reset ();
  mpu_init ();

  sb_write_memory (0x01, 0xC2);
  for ( i = stack_base - 0x10; i <= stack_base; ++i) sb_write_memory (i, 0x00);
  mpu_write (variant & 0x03);
  sb_delay (0xFFFF);
  sb_send (0xE3);

  i = sb_recv ();
  if (error != 0) return 0;
  if (i != 0xAA) return 0;
  return 1;
}

void stage1 (void) {
  uint8_t i, res;

  for (i=0; i < 8; ++i) {
    sprintf (strbuf, "Stage 1, variant %d ...", (int)i);
    log (strbuf);
    if (i < 4) res = stage1_inject1 (i);
    else res = stage1_inject2 (i);
    
    if (res) {
      sprintf (strbuf, "passed\n");
      log (strbuf);
    } else {
      sprintf (strbuf, "failed\n");
      log (strbuf);
    }
  }

}


uint16_t ADDR1 = 0x0A66;
const uint8_t stage1_result[8] = { 1, 0, 1, 0, 1, 0, 1, 1 };

uint8_t stage2_inject1 (uint8_t st1v, uint8_t st2v) {
  uint8_t i;
  uint8_t pc_ptr;
  
  error = 0;
  sb_reset ();
  mpu_init ();

  sb_write_memory (0x01, 0xC4);
  for ( i = stack_base - 0x10; i <= stack_base; ++i) sb_write_memory (i, 0x00);
  pc_ptr = stack_base - st2v;
  sb_write_memory (pc_ptr, (ADDR1 >> 8) & 0xFF);
  sb_write_memory (pc_ptr - 1, ADDR1 & 0xFF);
  mpu_write (st1v & 0x03);
  sb_delay (0xFFFF);

  i = sb_recv ();
  if (error != 0) return 0;
  if (i == 0xAA) return 0;
  return 1;
}

uint8_t stage2_inject2 (uint8_t st1v, uint8_t st2v) {
  uint8_t i;
  uint8_t pc_ptr;
  
  error = 0;
  sb_reset ();
  mpu_init ();

  sb_write_memory (0x01, 0xC2);
  for ( i = stack_base - 0x10; i <= stack_base; ++i) sb_write_memory (i, 0x00);
  pc_ptr = stack_base - st2v;
  sb_write_memory (pc_ptr, (ADDR1 >> 8) & 0xFF);
  sb_write_memory (pc_ptr - 1, ADDR1 & 0xFF);
  mpu_write (st1v & 0x03);
  sb_delay (0xFFFF);
  sb_send (0xE3);

  i = sb_recv ();
  if (error != 0) return 0;
  if (i == 0xAA) return 0;
  return 1;
}


void stage2 (void) {
  uint8_t i, res, st1v, st2v;

  for (i=0; i < 64; ++i) {
    
    st1v = i & 0x07;
    st2v = (i >> 3) & 0x07;

    sprintf (strbuf, "Stage 2, variant %02d ...", (int)i);
    log (strbuf);

    if (stage1_result[st1v] == 0) res = 2;
    else {
      if (st1v < 4) res = stage2_inject1 (st1v, st2v);
      else res = stage2_inject2 (st1v, st2v);
    }
    
    if (res == 1) {
      log ("passed\n");
    }
    if (res == 0) {
      log ("failed\n");
    }
    if (res == 2) { log ("skipped\n"); res = 0; }

  }
}

uint8_t stage3_inject2 () {
  uint8_t i;
  uint8_t pc_ptr;
  
  error = 0;
  sb_reset ();
  mpu_init ();

  sb_write_memory (0x01, 0xC2);
  sb_write_memory (0x66, 0xEB);
  for ( i = stack_base - 0x10; i <= stack_base; ++i) sb_write_memory (i, 0x00);
  pc_ptr = stack_base - 4;
  sb_write_memory (pc_ptr + 1, 0x66);
  sb_write_memory (pc_ptr, (ADDR1 >> 8) & 0xFF);
  sb_write_memory (pc_ptr - 1, ADDR1 & 0xFF);
  mpu_write ( 0x03);
  sb_delay (0xFFFF);
  sb_send (0xE3);

  i = sb_recv ();
  if (error != 0) return 0;
  if (i != 0xEB) return 0;
  return 1;
}


void stage3 (void) {
  uint8_t res;
  res = stage3_inject2 ();

  sprintf (strbuf, "Stage 3, variant 39 ...");
  log (strbuf);
  if (res == 1) {
    log ("passed\n");
  }
  if (res == 0) {
    log ("failed\n");
  }

}

uint8_t stage4_inject2 (uint16_t addr) {
  uint8_t i;
  uint8_t pc_ptr;
  
  error = 0;
  sb_reset ();
  mpu_init ();

  sb_write_memory (0x01, 0xC2);
  for ( i = stack_base - 0x10; i <= stack_base; ++i) sb_write_memory (i, 0x00);
  pc_ptr = stack_base - 4;
  sb_write_memory (pc_ptr, (addr >> 8) & 0xFF);
  sb_write_memory (pc_ptr - 1, addr & 0xFF);
  mpu_write ( 0x03);
  sb_delay (0xFFFF);
  sb_send (0xE1);

  i = sb_recv ();
  if (error != 0) return 0;
  if (i != 0x43) return 0;

  i = sb_recv ();
  if (error != 0) return 0;
  if (i != 0x4F) return 0;

  i = sb_recv ();
  if (error != 0) return 0;
  if (i != 0x50) return 0;

  i = sb_recv ();
  if (error != 0) return 0;
  if (i != 0x59) return 0;

  return 1;
}

void stage4 (void) {
  uint8_t res;
  uint16_t addr_l = 0x1180;
  uint16_t addr_h = 0x1181;

  for (;;) {
    sprintf (strbuf, "Stage 4, addr %04X ...", addr_l);
    log (strbuf);
    res = stage4_inject2 (addr_l);
    if (res) sprintf (strbuf, "passed\n");
    sprintf (strbuf, "failed\n");
    log (strbuf);
//    if (res) break;
    --addr_l;
    
    log (strbuf);
    sprintf (strbuf, "Stage 4, addr %04X ...", addr_h);
    log (strbuf);
    res = stage4_inject2 (addr_h);
    if (res) sprintf (strbuf, "passed\n");
    sprintf (strbuf, "failed\n");
    log (strbuf);
    ++addr_h;

    if (addr_h >= 0x1800) return;
  }
  sprintf (strbuf, "passed\n");
}

uint8_t stage5_inject2 (uint16_t addr) {
  uint8_t i;
  uint8_t pc_ptr;
  
  error = 0;
  sb_reset ();
  mpu_init ();

  sb_write_memory (0x01, 0xC2);
  for ( i = stack_base - 0x10; i <= stack_base; ++i) sb_write_memory (i, 0x00);
  pc_ptr = stack_base - 4;
  sb_write_memory (pc_ptr, (addr >> 8) & 0xFF);
  sb_write_memory (pc_ptr - 1, addr & 0xFF);
  mpu_write ( 0x03);
  sb_delay (0xFFFF);
  sb_send (0xE1);

  i = sb_recv ();
  if (error != 0) return 0;
  if (i != 0x02) return 0;

  return 1;
}


void stage5 (void) {
  uint8_t res;
  uint16_t addr_l = 0x1180;
  uint16_t addr_h = 0x1181;

  for (;;) {
    sprintf (strbuf, "Stage 5, addr %04X ...", addr_l);
    log (strbuf);
    res = stage5_inject2 (addr_l);
    if (res) sprintf (strbuf, "passed\n");
    sprintf (strbuf, "failed\n");
    log (strbuf);
//    if (res) break;
    --addr_l;
    
    log (strbuf);
    sprintf (strbuf, "Stage 5, addr %04X ...", addr_h);
    log (strbuf);
    res = stage5_inject2 (addr_h);
    if (res) sprintf (strbuf, "passed\n");
    sprintf (strbuf, "failed\n");
    log (strbuf);
    ++addr_h;

    if (addr_h >= 0x1800) return;
  }
  sprintf (strbuf, "passed\n");
}


void hack (void) {
  if (dsp_ver_l < 10) stack_base = 0x80;
  else stack_base = 0xC0;

  // print_codepoints_addr();
  // stage1 ();
  // stage2 ();
  // stage3 ();
  // stage4 ();
  stage5 ();
 

}


void main (int argc, char *argv[]) {
  

  error = 0;
  log_open ();
  sb_initial_diag ();

  if (error == 0) {
    hack ();
  }


  error_print ();
  log_close ();

  mpu_command (0xFF);
  mpu_read();
  sb_reset ();
}
